import os
from PIL import Image

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as transforms

import warnings; warnings.filterwarnings('ignore')

from .nets import Backbone
from ..utils import check_ckpt_exist, convert_image_type

class IdExtractor(nn.Module):
    def __init__(self, x=32, y=32, w=192, h=192 , root='~/.invz_packages/id_extractor', ckpt_name = 'currface.pth', url_id = '15xnMCs8udpODpSGwbePuhJqDoDeE0SA7', force=False):
        """
        Methods
        --------
        forward(tesnor_img : tensor) -> id_vector : tensor
        
        data_preprocess(img_path : str) -> tensor_img : tensor
        
        compare_similarity(id_vector_1 : tensor, id_vector_2 : tensor) -> score : float
        """
        super(IdExtractor, self).__init__()
        device = "cuda" if torch.cuda.is_available() else "cpu"
        self.id_extractor = Backbone().to(device).eval()
        
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=device)

        self.id_extractor.load_state_dict(ckpt)
        for param in self.id_extractor.parameters():
            param.requires_grad = False
        del ckpt

        self.x, self.y, self.w, self.h = x, y, w, h

        self.transform = transforms.Compose([
            transforms.Resize((256, 256), interpolation=Image.BILINEAR),
            transforms.ToTensor(),
            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
        ])
        
    def data_preprocess(self, input):
        """
        
        """
        pil_img = convert_image_type(input)
        tensor_img = self.transform(pil_img).unsqueeze(0).cuda()
        return tensor_img
        
    def forward(self, tensor_img):
        """
        
        """
        tensor_img = F.interpolate(tensor_img[..., self.y:self.y+self.h, self.x:self.x+self.w], (112, 112), mode='bilinear')
        id_vector = self.id_extractor(tensor_img)
        return id_vector
    
    def compare_similarity(self, id1, id2):
        """
        
        """
        score = torch.cosine_similarity(id1, id2, dim=1).mean().item()
        return score
    



import numpy as np

from .utils import align_face
from ..lmk_detector import LmkDetector
from ..utils import convert_image_type

import warnings; warnings.filterwarnings('ignore')

class FaceAligner:
    def __init__(self, size=1024):
        self.LD = LmkDetector()
        self.size = size

    def data_preprocess(self, input):
        pil_img = convert_image_type(input, target='pillow')
        _pil_img = np.array(pil_img)
        return _pil_img

    def data_postprocess(self, inputs):
        results = []
        for input in inputs:
            results.append(input[:,:,::-1])
        return results

    def get_align_face(self, image, lmk=None):
        if lmk is None:
            faces_count, faces_lmks, faces_bool = self.LD.get_landmark(image, get_5pts=True)
            if faces_count == 0:
                raise RuntimeError("There isn't exist any face in image!!")
            
            else:
                alinged_faces, trans_invs = [], []
                for idx in range(faces_count):
                    face_lmks = faces_lmks[idx]
                    aligned_face, trans_inv= align_face(image, face_lmks, self.size)
                    alinged_faces.append(aligned_face)
                    trans_invs.append(trans_inv)
                    
                return alinged_faces, trans_invs
            
        elif len(lmk) ==  5:
            aligned_face, trans_inv= align_face(image, lmk, self.size)
            
            return [aligned_face], [trans_inv]
            
        elif len(lmk) == 106:
            lmk = np.array([lmk[38], lmk[88], lmk[86], lmk[52], lmk[61]])
            aligned_face, trans_inv= align_face(image, lmk, self.size)
            
            return [aligned_face], [trans_inv]



import torch
import torch.nn.functional as F

def arrange_mask(parsing, output_size=512):
    Lbrow_mask = torch.where(parsing==2, 1, 0) + torch.where(parsing==3, 1, 0)
    Lbrow_mask[:, :, :, :output_size//2] = 0
    parsing = Lbrow_mask * 2 + (1 - Lbrow_mask) * parsing

    Rbrow_mask = torch.where(parsing==2, 1, 0) + torch.where(parsing==3, 1, 0)
    Rbrow_mask[:, :, :, output_size//2:] = 0
    parsing = Rbrow_mask * 3 + (1 - Rbrow_mask) * parsing

    Leye_mask = torch.where(parsing==4, 1, 0) + torch.where(parsing==5, 1, 0)
    Leye_mask[:, :, :, :output_size//2] = 0
    parsing = Leye_mask * 4 + (1 - Leye_mask) * parsing
    
    Reye_mask = torch.where(parsing==4, 1, 0) + torch.where(parsing==5, 1, 0)
    Reye_mask[:, :, :, output_size//2:] = 0
    parsing = Reye_mask * 5 + (1 - Reye_mask) * parsing
    
    Leye_mask = torch.where(parsing==7, 1, 0) + torch.where(parsing==8, 1, 0)
    Leye_mask[:, :, :, :output_size//2] = 0
    parsing = Leye_mask * 7 + (1 - Leye_mask) * parsing
    
    # Reye
    Reye_mask = torch.where(parsing==7, 1, 0) + torch.where(parsing==8, 1, 0)
    Reye_mask[:, :, :, output_size//2:] = 0
    parsing = Reye_mask * 8 + (1 - Reye_mask) * parsing
    return parsing

def get_one_hot(mask): # 0 ~ 8 h w
    mask_ = torch.tensor(mask, dtype=torch.int64)
    c = 19
    _,_,h,w = mask_.size()

    mask_ = torch.reshape(mask_,(1,1,h,w))
    one_hot_mask = torch.zeros(1, c, h, w)
    one_hot_mask_ = one_hot_mask.scatter_(1, mask_, 1.0)
    return one_hot_mask_



import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms

from .model import BiSeNet
from .utils import arrange_mask, get_one_hot
from ..utils import check_ckpt_exist, convert_image_type

# index = [
# 1 'skin', 2 'l_brow', 3 'r_brow', 4 'l_eye', 5 'r_eye', 6 'eye_g', 
# 7 'l_ear', 8 'r_ear', 9 'ear_r', 10 'nose', 
# 11 'mouth', 12 'u_lip', 13 'l_lip', 
# 14 'neck', 15 'neck_l', 16 'cloth', 17 'hair', 18 'hat']

class FaceParser(nn.Module):
    def __init__(self, root='~/.invz_packages/face_parser', ckpt_name = 'faceparser.pth', url_id = '1h5CG5VftFVtvS59fyZQzetlK_Nmyrjhj', force=False):
        super(FaceParser, self).__init__()
        
        device = "cuda" if torch.cuda.is_available() else "cpu"
        self.parsing_net = BiSeNet(n_classes=19).to(device)
        
        ckpt_path = check_ckpt_exist(root, ckpt_name = ckpt_name, url_id = url_id, force = force)
        ckpt = torch.load(ckpt_path, map_location=device)
        
        self.parsing_net.load_state_dict(ckpt)
        for param in self.parsing_net.parameters():
            param.requires_grad = False
        self.parsing_net.eval()
        del ckpt

        self.GaussianBlur = transforms.GaussianBlur(kernel_size=5, sigma=(0.1, 5))
        self.kernel = torch.ones((1,1,5,5), device="cuda")
        
        self.transform = transforms.Compose([
            transforms.Resize((512,512)),
            transforms.ToTensor(),
            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
        ])

    def data_preprocess(self, input):
        pil_img = convert_image_type(input)
        tensor_img = self.transform(pil_img).unsqueeze(0).cuda()
        return tensor_img
        
    def data_postprocess(self, tensor_label):
        cv2_label = tensor_label.squeeze().cpu().numpy()
        return cv2_label
    
    def get_label(self, tensor_img, size=512):
        label = self.parsing_net(tensor_img)
        _label = F.interpolate(label, (size,size), mode='bilinear').max(1)[1]
        _label = arrange_mask(_label, size)
        return _label
    
    def get_onehot(self, tensor_img, size=512):
        label = self.get_label(tensor_img, size)
        onehot = get_one_hot(label)
        return onehot


import torch.nn.functional as F
import numpy as np
import warnings; warnings.filterwarnings('ignore')
from insightface.app import FaceAnalysis # pip install -U insightface

from ..utils import convert_image_type

class LmkDetector:
    def __init__(self,):
        self.app = FaceAnalysis(providers=['CUDAExecutionProvider'], allowed_modules=['detection', 'landmark_2d_106'])
        self.app.prepare(ctx_id=0, det_size=(640, 640))

    def data_preprocess(self, input):
        pil_img = convert_image_type(input, target='pillow')
        return np.array(pil_img)

    def data_postprocess(self, inputs):
        results = []
        for input in inputs:
            results.append(input[:,:,::-1])
        return results

    def get_landmark(self, image, get_5pts=True):
        W, H = image.shape[:2]
        faces = self.app.get(np.pad(image, ((W//4, W//4), (H//4, H//4), (0, 0))))
        
        faces_count = len(faces)
        faces_lmks, faces_bool = [], []
        if len(faces):
            for idx in range(faces_count):
                lmk = faces[idx].landmark_2d_106 - [[H//4, W//4]]
            
                if get_5pts:
                    lmk = np.array([lmk[38], lmk[88], lmk[86], lmk[52], lmk[61]])
                    
                faces_bool.append(True)
                faces_lmks.append(lmk)
                
            return faces_count, faces_lmks, faces_bool
        
        elif not len(faces):
            return 0, None, None


import torch
import torch.nn.functional as F
from torchvision import transforms

def get_innerface_mask(label):
    innerface_mask  = torch.where(label<14, 1, 0)
    innerface_mask -= torch.where(label==0, 1, 0)
    
    return innerface_mask

def get_dilate_blur_mask(mask, do_dilate=True, do_blur=True, dilate_kernel_size=5, blur_kernel_size=5):
    assert len(mask.shape) != 4, "Mask must have batch-channel and tensor type."
    
    if do_dilate:
        kernel = torch.ones((1,1,dilate_kernel_size, dilate_kernel_size), device = mask.device)
        mask = torch.clamp(F.conv2d(mask, kernel, padding=(dilate_kernel_size//2, dilate_kernel_size//2)), 0, 1)

    if do_blur:
        GaussianBlur = transforms.GaussianBlur(kernel_size=blur_kernel_size, sigma=(0.1, 5))
        mask = GaussianBlur(mask)
        
    return mask


import os
import gdown

def check_ckpt_exist(root, ckpt_name, url_id, force):
    _root = os.path.expanduser(root)
    if not os.path.exists(_root):
            os.makedirs(_root)
            
    dir_path = os.path.join(_root, ckpt_name)
    if force:
        print('download_path :', dir_path)
        download_file(save_path = dir_path, url_id = url_id)
        return dir_path
        
    elif os.path.exists(dir_path):
        return dir_path
    
    else:
        print('download_path :', dir_path)
        download_file(save_path = dir_path, url_id = url_id)
        return dir_path

def download_file(save_path, url_id):
    """
    https://github.com/wkentaro/gdown
    """
    url = 'https://drive.google.com/u/0/uc?id=' + url_id + '&export=download'
    gdown.download(url, save_path, quiet=False)
        